import sys
sys.path.append("..")

from TokenManager import *
from AuthRepository import *

from FlaskServer import *
from Server import *
from Config import *
from RoleManager import *
from auth_dtos import *


def login(ctx: QRContext):
    login = ctx.json_data['login']
    password = ctx.json_data['password']
    user_id = ctx.repository.check_credentials(login, password)
    if user_id is None:
        return MethodResult('wrong credentials', 500)

    user = ctx.repository.get_user_data(user_id)
    if user is None:
        return MethodResult('account not found', 500)

    jwt_token = ctx.managers['token_manager'].make_token(user_id)
    return MethodResult(JwtDTO(jwt_token))


@require_token()
def user_info(ctx: QRContext, user_id):
    user = ctx.repository.get_user_data(user_id)
    if user is None:
        return MethodResult('account not found', 500)

    return MethodResult(UserInfoDTO(**user))


class AuthServer(FlaskServer, AuthRepository):
    """DI class"""


if __name__ == "__main__":
    config = QRYamlConfig()
    config.read_config('config.yaml')

    host = config['app']['host']
    port = config['app']['port']

    token_man = JwtTokenManager()
    token_man.load_config(config['jwt'])

    server = AuthServer()
    server.init_server(config['app'])
    if config['app']['logging']:
        server.configure_logger(config['app']['logging'])
    server.register_manager(token_man)

    server.register_method('/login', login, 'POST')
    server.register_method('/info', user_info, 'GET')
    server.run(host, port)



from .FileManager import *
from .FlaskServer import *
from .Repository import *
from .Server import *
from .Config import *
from .Logger import *
from .RoleManager import *
from .TokenManager import *
from .dict_parsing import *
from .dto_converter import *

__all__ = ['dict_parsing', 'dto_converter', 'FileManager', 'FlaskServer',
           'QRRepository', 'Server', 'Config', 'QRLogger', 'RoleManager', 'TokenManager']


import logging
from abc import abstractmethod, abstractproperty
from .Config import *

# todo manage logging to different places

class IQRLogger:
    @abstractmethod
    def configure_logger(self, config: IQRConfig):
        """configure logger"""

    @abstractmethod
    def info(self, msg, *args):
        """log info"""

    @abstractmethod
    def warning(self, msg, *args):
        """log warning"""

    @abstractmethod
    def error(self, msg, *args):
        """log error"""

    @abstractmethod
    def exception(self, msg, *args):
        """log exception"""


class QRLogger(IQRLogger):
    def __init__(self):
        self.logger = logging.getLogger()

    # todo log error here
    def configure_logger(self, config: IQRConfig):
        logger = logging.getLogger(config['logger_name'])
        logger.addHandler(get_stream_handler(config['level'].upper()))
        if config['file']:
            if config['file_level'] is None: file_level = config['level']
            else: file_level = config['file_level']
            logger.addHandler(get_file_handler(config['file'], file_level.upper()))

        logger = logging.LoggerAdapter(logger, {'app': config['app_name']})
        logger.setLevel('INFO')
        self.logger = logger

    def info(self, msg, *args):
        self.logger.info(msg, *args)

    @abstractmethod
    def warning(self, msg, *args):
        self.logger.warning(msg, *args)

    @abstractmethod
    def error(self, msg, *args):
        self.logger.error(msg, *args, exc_info=False)

    @abstractmethod
    def exception(self, msg, *args):
        self.logger.exception(msg, *args)


#(%(filename)s).%(funcName)s(%(lineno)d)
_log_format = "%(asctime)s [%(levelname)s]: [%(app)s, %(name)s] - %(message)s"

class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""
    grey = "\033[30m"
    yellow = "\033[33m"
    red = "\033[31m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    FORMATS = {
        logging.DEBUG: grey + _log_format + reset,
        logging.INFO: grey + _log_format + reset,
        logging.WARNING: yellow + _log_format + reset,
        logging.ERROR: red + _log_format + reset,
        logging.CRITICAL: bold_red + _log_format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_file_handler(filename, level):
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(level)
    file_handler.setFormatter(logging.Formatter(_log_format))
    return file_handler


def get_stream_handler(level):
    sh = logging.StreamHandler()
    sh.setLevel(level)
    sh.setFormatter(logging.Formatter(_log_format))
    sh.setFormatter(CustomFormatter())
    return sh



from abc import abstractmethod
from typing import List

from dataclasses import dataclass

class QRDTO:
    @abstractmethod
    def to_dict(self):
        return self.__dict__.copy()


def OneOfQRDTO(*variants):
    class OneOfDTO:
        __variants = variants
        def __init__(self, *args, **kwargs):
            self.obj = None

            for variant in self.__variants:
                try:
                    self.obj = variant(*args, **kwargs)
                    break
                except Exception as e:
                    continue
            if self.obj is None:
                raise Exception('OneOfDTO: no variants fit the data!')

        def to_dict(self):
            return self.obj.to_dict()
    return OneOfDTO


def ArrayQRDTO(objects_type):
    class ArrayDTO(QRDTO):
        def __init__(self, data: list):
            self.data = data
            for i in range(len(data)):
                if isinstance(data[i], dict):
                    self.data[i] = objects_type(**data[i])
                else:
                    self.data[i] = objects_type(data[i])

        def to_dict(self):
            res = self.data.copy()
            for i in range(len(res)):
                res[i] = res[i].to_dict()
            return res
    return ArrayDTO

@dataclass
class DefaultResponseDTO(QRDTO):
    response: str = 'ok'

    def to_dict(self):
        return self.response


def convert_fields(data: dict):
    def wrapper(dto_class: QRDTO):
        init = dto_class.__init__

        def __init__(self, *args, **kwargs):
            init(self, *args, **kwargs)
            for k, v in data.items():
                as_list = False
                if k.startswith('[]'):
                    k = k[2:]
                    as_list = True
                if self.__dict__.get(k) is not None:
                    stor = self.__dict__
                    if not as_list:
                        stor[k] = v(**stor[k])
                    else:
                        stor[k] = [v(**x) for x in stor[k]]

        dto_class.__init__ = __init__
        return dto_class
    return wrapper


def dto_kwargs_parser(func):
    def wrapper(dto_class: QRDTO):
        init = dto_class.__init__

        def __init__(self, *args, **kwargs):
            func(kwargs)
            init(self, *args, **kwargs)

        dto_class.__init__ = __init__
        return dto_class
    return wrapper


import Repository as rep

users = [('user1', 'pass1'), ('user2', 'pass2')]


class AuthRepository(rep.IQRRepository):
    def __init__(self):
        super().__init__()

    def connect_repository(self, configuration):
        pass

    def set_role(self, role: str):
        pass

    def check_credentials(self, login, password):
        for i, u in enumerate(users):
            if u == (login, password):
                return i
        return None

    def get_user_data(self, user_id):
        if user_id < len(users):
            data = users[user_id]
            return {'name': data[0], 'id': user_id}
        return None


from dict_parsing import *
from dto_converter import *


def user_info_parser(d: dict):
    parse_dict(d, rename={'surname': 'last_name'})


@dataclass
class JwtDTO(QRDTO):
    access_token: str


@dto_kwargs_parser(user_info_parser)
@dataclass
class UserInfoDTO(QRDTO):
    name: str
    id: int



from flask import Flask, jsonify
from flask_cors import CORS
from flask import request, Response
import time
import gevent.pywsgi

from .Server import *


class FlaskContextCreator(IQRContextCreator):
    def create_context(self, request, rep: IQRRepository, meta: dict) -> QRContext:
        json = request.get_json(silent=True)
        if json is None:
            json = dict()
        return QRContext(json, request.args, request.headers, request.form, request.files, rep, meta)


class FlaskServer(IQRServer, FlaskContextCreator, QRLogger):
    def __init__(self, default_err_code=500, default_err_msg='Internal server error'):
        FlaskContextCreator.__init__(self)
        QRLogger.__init__(self)
        self.app = None
        self.debug = None
        self.methods = {}
        self.managers = dict()
        self.meta = dict()

        self.default_err_code = default_err_code
        self.default_err_msg = default_err_msg

    def init_server(self, config: IQRConfig):
        app_name = config['app_name']
        if app_name is None: app_name = 'app'

        debug = config['debug']
        if debug is None: debug = False

        self.app = Flask(app_name)
        CORS(self.app)
        self.debug = debug

    def run(self, host, port):
        self.logger.info('running server on %s:%s' % (host, port))
        self.app_server = gevent.pywsgi.WSGIServer((host, port), self.app)
        self.app_server.serve_forever()
        # self.app.run(debug=self.debug, host=host, port=port)

    def register_method(self, route: str, f, method_type: str):
        """register method"""
        func = lambda *args, **kwargs: self.__method(f, *args, **kwargs)
        func.__name__ = f.__name__
        self.methods[f.__name__] = \
            self.app.route(route, methods=[method_type])(func)

    def __method(self, f, *args, **kwargs):
        ctx = super().create_context(request, self, meta=self.meta)
        ctx.set_managers(self.managers)
        in_msg = '[' + request.method + '] ' + request.url + '/' + request.query_string.decode()
        try:
            start = time.time()
            result = f(ctx, *args, **kwargs)
            end = time.time()
            msecs = int((end - start) * 1000)
            super().info('[' + str(msecs) + ' msecs]' + in_msg)

            if result.raw_data:
                resp = Response(result.result, result.status_code)
            else:
                resp = jsonify(result.result)
                resp.status_code = result.status_code

            if result.headers is not None:
                for header, value in result.headers.items():
                    resp.headers[header] = value
            return resp

        except Exception as e:
            super().info(in_msg)
            super().exception(e)
            return self.default_err_msg, self.default_err_code

    def register_manager(self, manager: IQRManager):
        self.managers[manager.get_name()] = manager

    def set_meta(self, meta: dict):
        self.meta = meta


from abc import abstractmethod, abstractproperty
from .Repository import *
from .Logger import *
from .Config import *
from .dto_converter import *

class IQRManager:
    """abstract manager class"""
    @staticmethod
    @abstractmethod
    def get_name() -> str:
        """how the manager is named"""


class QRContext:
    """object representing the request context, providing wrap around request's body and headers
    as well as repository instance and other possible managers"""

    def __init__(self, json_data, params, headers, form, files, repository: IQRRepository, meta=None):
        self.__json_data = json_data
        self.__params = params
        self.__headers = headers
        self.__rep = repository
        self.__form = form
        self.__files = files
        self.managers = dict()

        if meta is None:
            meta = dict()
        self.meta = meta

    def get_json_data(self): return self.__json_data

    def get_params(self): return self.__params

    def get_headers(self): return self.__headers

    def get_form(self): return self.__form

    def get_files(self): return self.__files

    def get_repository(self): return self.__rep

    def get_manager(self, name):
        return self.managers.get(name)

    params = property(get_params)
    json_data = property(get_json_data)
    headers = property(get_headers)
    form = property(get_form)
    files = property(get_files)
    repository = property(get_repository)

    def add_manager(self, name, manager: IQRManager):
        self.managers[name] = manager

    def set_managers(self, managers):
        self.managers = managers


class IQRContextCreator:
    @abstractmethod
    def create_context(self, *args) -> QRContext:
        """create request's context using given data"""


class MethodResult:
    def __init__(self, result=None, status_code=None, raw_data=False, headers=None):
        if status_code is None: status_code = 200
        self.result = result
        if isinstance(result, QRDTO):
            self.result = self.result.to_dict()

        self.status_code = status_code
        self.raw_data = raw_data
        self.headers = headers


class IQRServer(IQRRepository, IQRLogger, IQRContextCreator):
    @abstractmethod
    def init_server(self, configuration: IQRConfig):
        """initialize server"""

    @abstractmethod
    def run(self, host, port):
        """run the server"""

    @abstractmethod
    def register_method(self, route: str, f, method_type: str):
        """
        register method. Server provides positional QRContext argument and transfers all other *arg and **kwarg arguments
        :param method_type: one of 'GET', 'POST', 'PUT'
        """

    @abstractmethod
    def register_manager(self, manager: IQRManager):
        """
        register manager
        :param manager: the manager itself
        :return:
        """



import confuse
import pathlib
from collections import OrderedDict
from abc import abstractmethod
import distutils.util as util
import os

class IQRConfig:
    """abstract class for config"""

    @abstractmethod
    def read_config(self, *args):
        """read config and make fields available either as 'get' methods or as object's properties"""

    @abstractmethod
    def get(self, key: str):
        """return value of key. if value is a hierarchy itself, IQRConfig object is returned"""

    @abstractmethod
    def __getitem__(self, key: str):
        """return value of key. if value is a hierarchy itself, IQRConfig object is returned"""

class QRYamlConfig(IQRConfig):
    def __init__(self):
        self.data = dict()

    def __getitem__(self, key):
        return self.data.get(key)

    def __setitem__(self, key, value):
        self.data[key] = value
        self.__dict__[key] = value

    def get(self, key):
        return self.data.get(key)

    def read_config(self, config_name='config.yaml', directory=None):
        if directory is None:
            #directory = str(pathlib.Path(__file__).parent.absolute())
            directory = os.getcwd()
        config = confuse.Configuration('app')
        config.set_file(directory + '/' + config_name)

        for x in config:
            d = config[x].get()
            self.data[x] = self.__parse_dict(d)
            self.__dict__[x] = self.data[x]

    def __parse_dict(self, d):
        if not isinstance(d, OrderedDict):
            # d is string here
            # if isinstance(d, bool):
            #     return d
            # elif isinstance(d, int):
            #     return d
            # elif d.replace('.', '', 1).isdigit():
            #     return float(d)
            return d

        obj = QRYamlConfig()
        for k, v in d.items():
            obj.data[k] = self.__parse_dict(v)
            obj.__dict__[k] = obj.data[k]
        return obj



import setuptools


required = [
    'qrookDB',
    'Flask == 2.0.1',
    'Flask-HTTPAuth == 4.2.0',
    'Flask-Cors == 3.0.10',
    'Jinja2==3.0.1',
    'confuse',
    'PyJWT',
    'psycopg2-binary == 2.8.6',
    'Inject',
    'requests',
    'gevent',
]

setuptools.setup(
    name="qr_server",
    version="1.1.11",
    author="Kurush",
    author_email="ze17@ya.ru",
    description="console app builder",
    long_description_content_type="text/markdown",
    url="https://github.com/Kurush7/qr_server",
    packages=setuptools.find_packages(),
    install_requires=required,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    long_description='''
# qr_server

This project is the extension of Flask project (https://pypi.org/project/Flask/), aimed on fast creation of web-apps with minimalistic syntax.
The solution provides support for basic HTTP-routing (with file sending), fast DTOs 
(data transfer objects for formalization and validation of response data),
simple jwt-token system, role manager (database-side rights system) and configurable logging.
See 'example' directory for a minimal working application built using this library

### Usage example:
```python
from qr_server.Server import MethodResult, QRContext
from qr_server.Config import QRYamlConfig
from qr_server.TokenManager import require_token, JwtTokenManager
from qr_server.FlaskServer import FlaskServer

def login(ctx: QRContext):
    login = ctx.json_data['login']
    password = ctx.json_data['password']
    user_id = ctx.repository.check_credentials(login, password)
    if user_id is None:
        return MethodResult('wrong credentials', 500)

    user = ctx.repository.get_user_data(user_id)
    if user is None:
        return MethodResult('account not found', 500)

    jwt_token = ctx.managers['token_manager'].make_token(user_id)
    return MethodResult(JwtDTO(jwt_token))


@require_token()
def user_info(ctx: QRContext, user_id):
    user = ctx.repository.get_user_data(user_id)
    if user is None:
        return MethodResult('account not found', 500)

    return MethodResult(UserInfoDTO(**user))


class AuthServer(FlaskServer, AuthRepository):
    """DI class"""


if __name__ == "__main__":
    config = QRYamlConfig()
    config.read_config('config.yaml')

    host = config['app']['host']
    port = config['app']['port']

    token_man = JwtTokenManager()
    token_man.load_config(config['jwt'])

    server = AuthServer()
    server.init_server(config['app'])
    if config['app']['logging']:
        server.configure_logger(config['app']['logging'])
    server.register_manager(token_man)

    server.register_method('/login', login, 'POST')
    server.register_method('/info', user_info, 'GET')
    server.run(host, port)
```
'''
)



from abc import abstractmethod, abstractproperty
from .Config import *
from qrookDB.DB import DB

class IQRRepository:
    """abstract class representing the Repository object - one providing data-managing functions"""
    @abstractmethod
    def connect_repository(self, configuration: IQRConfig):
        """initialize repository with configuration given"""

    @abstractmethod
    def set_role(self, role: str):
        """set database role for the request"""

class QRRepository(IQRRepository):
    def __init__(self):
        self.db = None

    def connect_repository(self, config: IQRConfig):
        conn = [config['connector'],
                config['dbname'],
                config['username'],
                config['password'],
                config['host'],
                config['port']
                ]
        self.db = DB(*conn, format_type='dict')
        self.db.create_data()
        if config['logging']:
            lg = config['logging']
            fl = config['level'] if lg['file_level'] is None else lg['file_level']
            fl = fl.upper()
            self.db.create_logger(lg['logger_name'], lg['app_name'], lg['level'].upper(), lg['file'], fl)

    def set_role(self, role: str = 'guest'):
        if not role.isidentifier():
            raise Exception('set_role method got not an identifier!')
        request = self.db.exec('set role "%s"' % role)
        request.exec()
        if request.get_error() != None:
            raise Exception('failed to set role: ', request.get_error())


def parse_dict(d: dict, remove: list = None, rename: dict = None, recursive=True):
    """changed 'd' param by removing fields in 'remove' and renaming ones in 'rename'"""
    if remove:
        for r in remove:
            if d.get(r) is None: continue
            d.pop(r)
    if rename:
        for k, v in rename.items():
            if d.get(k) is None: continue
            d[v] = d[k]
            d.pop(k)

    if recursive:
        for k in d.keys():
            if isinstance(d[k], dict):
                parse_dict(d[k], remove, rename, recursive)
            elif isinstance(d[k], list):
                for x in d[k]:
                    if isinstance(x, dict):
                        parse_dict(x, remove, rename, recursive)

    return d


def drop_none(d: dict):
    for k in list(d.keys()):
        if d[k] is None:
            d.pop(k)
    return d



from dataclasses import dataclass
import requests
import json

@dataclass
class QRAddress:
    host: str
    port: str

    def get_full_url(self, url):
        return f'{self.host}:{self.port}/{url}'


class QRRequest:
    def __init__(self, json_data=None, params=None, headers=None, form=None):
        self.__json_data = json_data
        self.__params = params
        self.__headers = headers
        self.__form = form

    def get_json_data(self): return self.__json_data

    def get_params(self): return self.__params

    def get_headers(self): return self.__headers

    def get_form(self): return dict(self.__form) if self.__form is not None else dict()

    params = property(get_params)
    json_data = property(get_json_data)
    headers = property(get_headers)

    def get_args(self) -> dict:
        args = {
            'json': self.get_json_data(),
            'params': self.get_params(),
            'headers': self.get_headers(),
            'data': self.get_form(),
        }
        for a in 'json', 'params', 'data':
            if len(args[a]) == 0:
                args.pop(a)
        if args['headers'].get('CONTENT_TYPE') is not None:
            # clear content type in case if 'data' (form) field is present - it needs boundaries which are better to be recalculated
            args['headers'] = dict(args['headers'])
            args['headers'].pop('Content-Type')
        return args


@dataclass
class QRResponse:
    ok: bool
    status_code: int
    reason: str
    content: bytes

    def get_json(self):
        return json.loads(self.content)


def send_request(address: QRAddress, url: str, method='GET', request: QRRequest = None):
    if request is None:
        request = QRRequest(None, None, None)

    resp = requests.request(method, address.get_full_url(url), **request.get_args())

    return QRResponse(resp.ok, resp.status_code, resp.reason, resp.content)


import datetime
import jwt
from abc import abstractmethod

from .Server import *

class ITokenManager(IQRManager):
    @staticmethod
    def get_name() -> str:
        return "token_manager"

    @abstractmethod
    def load_config(self, config: IQRConfig):
        """define tokenizing parameters"""

    @abstractmethod
    def make_token(self, user_id: int, payload: dict = None):
        """generate token with given payload (user_id is its part, required).
        adds 'exp' key - expired time defined by config"""

    @abstractmethod
    def decode_token(self, token, verify=True):
        """get token payload"""

    @abstractmethod
    def expired_exception(self):
        """exception to catch on token-expired occasion"""


class JwtTokenManager(ITokenManager):
    def __init__(self):
        self.secret = None
        self.algorithm = None
        self.exp = None

    @abstractmethod
    def expired_exception(self):
        return jwt.exceptions.ExpiredSignatureError

    def load_config(self, config: IQRConfig):
        self.secret = config['secret']
        self.algorithm = config['algorithm']
        self.exp = config['exp_seconds']

    def make_token(self, user_id: int, payload: dict = None):
        if payload is None: payload = dict()
        payload = payload.copy()
        payload['user_id'] = user_id
        payload['exp'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=self.exp)

        jwt_token = jwt.encode(payload, self.secret, self.algorithm)
        return jwt_token

    def decode_token(self, token, verify=True):
        payload = jwt.decode(token, self.secret, algorithms=[self.algorithm])
        return payload


def require_token(ignore_expired=False, send_token=False, send_db_role=False):
    def wrapper(f):
        def decorator(ctx: QRContext, *args, **kwargs):
            auth = ctx.headers.get('Authorization')
            if auth is None:
                return MethodResult('no auth data found', 401)
            if not auth.startswith('Bearer '):
                return MethodResult('Bearer token expected', 401)
            token = auth[len('Bearer '):]
            if send_token:
                kwargs['token'] = token
            try:
                payload = ctx.managers[ITokenManager.get_name()].decode_token(token, verify=True)
                if send_db_role:
                    kwargs['role'] = send_db_role
                return f(ctx, *args, **kwargs, user_id=payload['user_id'])
            except ctx.managers[ITokenManager.get_name()].expired_exception() as e:
                if ignore_expired:
                    payload = ctx.managers[ITokenManager.get_name()].decode_token(token, verify=False)
                    try:
                        return f(ctx, *args, **kwargs, user_id=payload['user_id'])
                    except Exception as e:
                        return MethodResult(str(e), 500)
                else:
                    return MethodResult('token expired', 401)
            except Exception as e:
                return MethodResult(str(e), 500)
        decorator.__name__ = f.__name__
        return decorator
    return wrapper


class MockTokenManager(ITokenManager):
    def load_config(self, config: IQRConfig):
        pass

    def make_token(self, user_id: int, payload: dict = None):
        return 'Bearer ' + str(user_id)

    @abstractmethod
    def decode_token(self, token, verify=True):
        if token is None or token == '':
            raise Exception('token expired')
        return {'user_id': int(token)}


from flask import send_from_directory
from abc import abstractmethod
import os.path

from .Server import *

class IFileManager(IQRManager):
    @staticmethod
    def get_name() -> str:
        return "file_manager"

    @abstractmethod
    def send_file(self, dirname, filename):
        """send file"""


class FlaskFileManager(IFileManager):
    def __init__(self):
        self.secret = None
        self.algorithm = None
        self.exp = None

    def send_file(self, dirname, filename):
        return send_from_directory(dirname, filename)


class MockFileManager(IFileManager):
    def send_file(self, dirname, filename):
        if filename in ['a', 'b', 'c']: return filename
        return 'file not found'


from .Server import QRContext


def role_manager(default_role='guest'):
    def decorator(f):
        def wrapper(ctx: QRContext, *args, **kwargs):
            role = kwargs.get('role')
            if role is None:
                role = default_role
            else:
                kwargs.pop('role')

            ctx.repository.set_role(role)
            return f(ctx, *args, **kwargs)
        wrapper.__name__ = f.__name__
        return wrapper

    return decorator